/*
 * uio_access.cpp
 * >  How to access a memory map via UIO driver
 *
 * Author: Ludwig Zenz (lzenz@dh-electronics.de)
 * Date: 2015-05-27
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>

struct uio_map_info {
	struct uio_map_info* prev;
	unsigned long 	offset;
	int				size;
	char*			name;
	int				number;
	char*			dev;
	uio_map_info(struct uio_map_info* prev, unsigned long offset, int size, char* name, int number, char* dev) :
                        prev(prev),
                        offset(offset),
                        size(size),
                        name(name),
                        number(number),
                        dev(dev) {}
};


static struct uio_map_info main_map0(NULL, 0x00000000, 0x04000000, 
                                        (char*)"uio_cs0", 0, (char*)"/dev/uio0");
/* 
 * example for a second map 
 * (has to match the device tree configuration) 
 */
/* static struct uio_map_info main_map1(&main_map0, 0x02000000, 0x02000000, 
                                        (char*)"uio_cs0_second_map", 1, (char*)"/dev/uio0");
 */


/* Use mmap()'ped address "iomem", 
 * not physical MYFPGA address */
#define MYFPGA_REG(iomem, offset) (volatile uint16_t*)(iomem + offset) // Third 16-bit reg


int main (int argc, char *argv[])
{
        int fd;
        void* iomem;
        char c, do_write = 0;
        unsigned int addr, temp = 0;
        unsigned short value = 0;
        struct uio_map_info* temp_uio = NULL;
        char *pEnd;

        /* check arguments and options */
        if (argc < 2) {
		printf("To few arguments!! \n");
		c = 'h';
	} else if ((c = getopt(argc, argv, "rwh")) == -1) {
                printf("Unknown option -%c \n\n", c);
                c = 'h';
        }

        switch (c) {
	case 'r': /* reading */
                if (argc < 3) {
                        printf("To few arguments!! \n");
                        c = 'h';
                }

                printf( "\n--> reading: \n");
                do_write = 0;
                break;

	case 'w': /* writing */
                if (argc < 4) {
                        printf("To few arguments!! \n");
                        return -1;
		}

                printf( "\n--> writing: \n");
                do_write = 1;
                break;

	case 'h':
                printf( "------- Usage: ------- \n");
                printf( "reading: -r <addr> \n");
                printf( "writing: -w <addr> <value> \n");
                printf( "help:    -h \n");
                return 0;

	default:
                /* should never be executed */
                return -1;
	}

        addr = strtol(argv[2], &pEnd, 16);
        
        /* allow only even addresses 
           because we have 16 bit registers 
           -> if the address is not aligned two accesses are made */
        addr &= 0xFFFE;  

        /* search the matching uio mmap */
        temp_uio = &main_map0; /* &main_map1 if you have a second map */
        
        /* search within the chain of 'uio_map_info' until no prev one is available */
        while (temp_uio != NULL) { 
    	        if (addr >= temp_uio->offset && addr < temp_uio->offset + temp_uio->size)
    		        break;
    	        else
    		        temp_uio = temp_uio->prev;
        }

        if (temp_uio == NULL) {
                printf("given address is not valid: %4x\n", addr);
                return -1;
        }

        fd = open(temp_uio->dev, O_RDWR|O_SYNC);
        if (fd < 0) {
                printf("failed to open %s, quitting\n", temp_uio->dev);
                return -1;
        } else {
                printf("succeeded to open %s!\n", temp_uio->dev);
        }

        printf("using map: %s\n\n", temp_uio->name);


        /* The mapping will be created nearby a page boundary */
        /* Note offset has a special meaning with uio devices */
        iomem = mmap(	NULL, 
    		        temp_uio->size,
			PROT_READ | PROT_WRITE, 
			MAP_SHARED, 
			fd, 
			temp_uio->number * getpagesize()); /* specifies which memory 
			                                      map of uio has to be used */

        if (iomem == MAP_FAILED) {
                printf("mapping of %s failed, quitting\n", temp_uio->name);
                close(fd);
                return -1;
        }

        if (do_write) { /* writing */
                value = strtol(argv[3], &pEnd, 16);
                *MYFPGA_REG((char*) iomem, addr) = value;
                printf("Written Value: %x \t to Addr: %x \n", value, addr);
    	
        } else { /* reading */
                temp = *MYFPGA_REG((char*) iomem, addr);
                printf("Read Value: %x  \t from Addr: %x \n", temp, addr);
        }

        munmap(iomem, temp_uio->size);
        close(fd);
        return 0;
}

